import 'package:badges/badges.dart' as badges;
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:overlay_support/overlay_support.dart';

import 'package:sudoku/config/default_global_settings.dart';
import 'package:sudoku/cubit/game_cubit.dart';
import 'package:sudoku/models/game.dart';

class GlobalAppBar extends StatelessWidget implements PreferredSizeWidget {
  const GlobalAppBar({super.key});

  @override
  Widget build(BuildContext context) {
    return BlocBuilder<GameCubit, GameState>(
      builder: (BuildContext context, GameState gameState) {
        final Game game = gameState.game;

        final List<Widget> menuActions = [];

        if (game.isRunning && !game.isFinished) {
          final GameCubit gameCubit = BlocProvider.of<GameCubit>(context);

          menuActions.add(TextButton(
            child: Container(
              decoration: BoxDecoration(
                borderRadius: BorderRadius.circular(4),
                border: Border.all(
                  color: Colors.white,
                  width: 3,
                ),
              ),
              child: const Image(
                image: AssetImage('assets/icons/button_back.png'),
                fit: BoxFit.fill,
              ),
            ),
            onPressed: () => toast('Long press to quit game...'),
            onLongPress: () {
              gameCubit.quitGame();
            },
          ));
          menuActions.add(const Spacer(flex: 6));
          menuActions.add(TextButton(
            child: Container(
              decoration: BoxDecoration(
                borderRadius: BorderRadius.circular(4),
                border: Border.all(
                  color: Colors.white,
                  width: 3,
                ),
              ),
              child: badges.Badge(
                showBadge: game.givenTipsCount == 0 ? false : true,
                badgeStyle: badges.BadgeStyle(
                  badgeColor: game.givenTipsCount < 10
                      ? Colors.green
                      : game.givenTipsCount < 20
                          ? Colors.orange
                          : Colors.red,
                ),
                badgeContent: Text(
                  game.givenTipsCount == 0 ? '' : game.givenTipsCount.toString(),
                  style: const TextStyle(color: Colors.white),
                ),
                child: Container(
                  padding: EdgeInsets.all(15 *
                      game.buttonTipsCountdown /
                      DefaultGlobalSettings.defaultTipCountDownValueInSeconds),
                  child: const Image(
                    image: AssetImage('assets/icons/button_help.png'),
                    fit: BoxFit.fill,
                  ),
                ),
              ),
            ),
            onPressed: () {
              final GameCubit gameCubit = BlocProvider.of<GameCubit>(context);
              game.canGiveTip() ? game.showTip(gameCubit) : null;
            },
          ));
          menuActions.add(const Spacer());
          menuActions.add(TextButton(
            child: Container(
              decoration: BoxDecoration(
                borderRadius: BorderRadius.circular(4),
                border: Border.all(
                  color: game.showConflicts == true ? Colors.blue : Colors.white,
                  width: 3,
                ),
              ),
              child: const Image(
                image: AssetImage('assets/icons/button_show_conflicts.png'),
                fit: BoxFit.fill,
              ),
            ),
            onPressed: () {
              gameCubit.toggleShowConflicts();
            },
          ));
        }

        return AppBar(
          title: const SizedBox.shrink(),
          actions: menuActions,
        );
      },
    );
  }

  @override
  Size get preferredSize => const Size.fromHeight(50);
}
